# -*- coding: utf-8 -*-
import base64

from Crypto.Hash import SHA
from Crypto.PublicKey import RSA
from Crypto.Signature import PKCS1_v1_5
from OpenSSL import crypto


class Fiel():
    def __init__(self, cer_der, key_der, passphrase):
        self.__importar_cer__(cer_der)
        self.__importar_key__(key_der, passphrase)

    def __importar_cer__(self, cer_der):
        # Cargar certificado en formato DER
        self.cer = crypto.load_certificate(crypto.FILETYPE_ASN1, cer_der)

    def __importar_key__(self, key_der, passphrase):
        # Importar KEY en formato DER
        self.key = RSA.importKey(key_der, passphrase)
        # Crear objeto para firmar
        self.signer = PKCS1_v1_5.new(self.key)

    def firmar_sha1(self, texto):
        # Generar SHA1
        sha1 = SHA.new(texto)
        # Firmar
        firma = self.signer.sign(sha1)
        # Pasar a base64
        b64_firma = base64.b64encode(firma)
        return b64_firma

    def cer_to_base64(self):
        # Extraer DER de certificado
        cer = crypto.dump_certificate(crypto.FILETYPE_ASN1, self.cer)
        # Pasar a b64
        return base64.b64encode(cer)

    def cer_issuer(self):
        # Extraer issuer
        d = self.cer.get_issuer().get_components()
        # Generar cadena issuer
        return u','.join(['{key}={value}'.format(key=key.decode(), value=value.decode()) for key, value in d])

    def cer_serial_number(self):
        # Obtener numero de serie del certificado
        serial = self.cer.get_serial_number()
        # Pasar numero de serie a string
        return str(serial)



# -*- coding: utf-8 -*-
from .webservicerequest import WebServiceRequest


class DescargaMasiva(WebServiceRequest):

    xml_name = 'descargamasiva.xml'
    soap_url = 'https://cfdidescargamasiva.clouda.sat.gob.mx/DescargaMasivaService.svc'
    soap_action = 'http://DescargaMasivaTerceros.sat.gob.mx/IDescargaMasivaTercerosService/Descargar'
    solicitud_xpath = 's:Body/des:PeticionDescargaMasivaTercerosEntrada/des:peticionDescarga'
    result_xpath = 's:Body/RespuestaDescargaMasivaTercerosSalida/Paquete'

    def descargar_paquete(self, token, rfc_solicitante, id_paquete):

        arguments = {
            'RfcSolicitante': rfc_solicitante,
            'IdPaquete': id_paquete,
        }

        element_response = self.request(token, arguments)

        respuesta = element_response.getparent().getparent().getparent().find(
            's:Header/h:respuesta', namespaces=self.external_nsmap
        )

        ret_val = {
            'cod_estatus': respuesta.get('CodEstatus'),
            'mensaje': respuesta.get('Mensaje'),
            'paquete_b64': element_response.text,
        }

        return ret_val



# -*- coding: utf-8 -*-
from .webservicerequest import WebServiceRequest


class VerificaSolicitudDescarga(WebServiceRequest):

    xml_name = 'verificasolicituddescarga.xml'
    soap_url = 'https://cfdidescargamasivasolicitud.clouda.sat.gob.mx/VerificaSolicitudDescargaService.svc'
    soap_action = 'http://DescargaMasivaTerceros.sat.gob.mx/IVerificaSolicitudDescargaService/VerificaSolicitudDescarga'
    solicitud_xpath = 's:Body/des:VerificaSolicitudDescarga/des:solicitud'
    result_xpath = 's:Body/VerificaSolicitudDescargaResponse/VerificaSolicitudDescargaResult'

    def verificar_descarga(self, token, rfc_solicitante, id_solicitud):

        arguments = {
            'RfcSolicitante': rfc_solicitante,
            'IdSolicitud': id_solicitud,
        }

        element_response = self.request(token, arguments)

        ret_val = {
            'cod_estatus': element_response.get('CodEstatus'),
            'estado_solicitud': element_response.get('EstadoSolicitud'),
            'codigo_estado_solicitud': element_response.get('CodigoEstadoSolicitud'),
            'numero_cfdis': element_response.get('NumeroCFDIs'),
            'mensaje': element_response.get('Mensaje'),
            'paquetes': []
        }

        for id_paquete in element_response.iter('{{{}}}IdsPaquetes'.format(self.external_nsmap[''])):
            ret_val['paquetes'].append(id_paquete.text)

        return ret_val



import os

from lxml import etree


class Utils():

    internal_nsmap = {
        's': 'http://schemas.xmlsoap.org/soap/envelope/',
        'o': 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd',
        'u': 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd',
        'des': 'http://DescargaMasivaTerceros.sat.gob.mx',
        '': 'http://www.w3.org/2000/09/xmldsig#',
        #'': 'http://DescargaMasivaTerceros.gob.mx',
    }

    external_nsmap = {
        '': 'http://DescargaMasivaTerceros.sat.gob.mx',
        's': 'http://schemas.xmlsoap.org/soap/envelope/',
        'u': 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd',
        'o': 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd',
        'h':'http://DescargaMasivaTerceros.sat.gob.mx',
        'xsi': 'http://www.w3.org/2001/XMLSchema-instance',
        'xsd': 'http://www.w3.org/2001/XMLSchema',
    }

    xml_name: str = None

    def __init__(self) -> None:
        self.read_xml(self.xml_name)

    def read_xml(self, xml_name: str) -> etree.Element:
        current_dir = os.path.dirname(os.path.abspath(__file__))
        xml_path = os.path.join(current_dir, xml_name)
        parser = etree.XMLParser(remove_blank_text=True)
        self.element_root = etree.parse(xml_path, parser).getroot()
        return self.element_root

    def get_element(self, xpath: str) -> etree.Element:
        return self.element_root.find(xpath, self.internal_nsmap)

    def get_element_external(self, element: etree.Element, xpath: str) -> etree.Element:
        return element.find(xpath, self.external_nsmap)

    def set_element_text(self, xpath: str, text: str) -> None:
        element = self.element_root.find(xpath, self.internal_nsmap)
        element.text = text

    @classmethod
    def element_to_bytes(cls, element: etree.Element) -> bytes:
        return etree.tostring(element, method='c14n', exclusive=1)



import base64
import hashlib

from lxml import etree

from .fiel import Fiel
from .utils import Utils


class Signer(Utils):

    nsmap = {
        None: 'http://www.w3.org/2000/09/xmldsig#'
    }

    xml_name = 'signer.xml'

    def __init__(self, fiel: Fiel) -> None:
        super().__init__()
        self.fiel = fiel

    def sign(self, element: etree.Element) -> etree.Element:

        element_bytes = self.element_to_bytes(element.getparent())
        element_hash = hashlib.new('sha1', element_bytes)
        element_digest = element_hash.digest()
        element_digest_base64 = base64.b64encode(element_digest)

        digest_xpath = 'SignedInfo/Reference/DigestValue'
        digest_element = self.get_element(digest_xpath)
        digest_element.text = element_digest_base64

        signed_info_xpath = 'SignedInfo'
        signed_info = self.get_element(signed_info_xpath)
        signed_info_bytes = self.element_to_bytes(signed_info)
        signed_info_sign = self.fiel.firmar_sha1(signed_info_bytes)

        xpath = 'SignatureValue'
        self.set_element_text(xpath, signed_info_sign)

        xpath = 'KeyInfo/X509Data/X509Certificate'
        self.set_element_text(xpath, self.fiel.cer_to_base64())

        xpath = 'KeyInfo/X509Data/X509IssuerSerial/X509IssuerName'
        self.set_element_text(xpath, self.fiel.cer_issuer())

        xpath = 'KeyInfo/X509Data/X509IssuerSerial/X509SerialNumber'
        self.set_element_text(xpath, self.fiel.cer_serial_number())

        element.append(self.element_root)

        return element



# -*- coding: utf-8 -*-
from .autenticacion import Autenticacion
from .descargamasiva import DescargaMasiva
from .fiel import Fiel
from .solicitadescarga import SolicitaDescarga
from .validacioncfdi import Validacion
from .verificasolicituddescarga import VerificaSolicitudDescarga

__all__ = [
    'Autenticacion',
    'DescargaMasiva',
    'Fiel',
    'SolicitaDescarga',
    'Validacion',
    'VerificaSolicitudDescarga',
]

name = 'cfdiclient'

version = '1.5.4'



# -*- coding: utf-8 -*-
import requests
from lxml import etree


class Validacion():
  def __init__(self):
    self.SOAP_URL = 'https://consultaqr.facturaelectronica.sat.gob.mx/ConsultaCFDIService.svc'
    self.SOAP_ACTION = 'http://tempuri.org/IConsultaCFDIService/Consulta'
    self.NSMAP = {
        's': 'http://schemas.xmlsoap.org/soap/envelope/',
        'des': 'http://DescargaMasivaTerceros.sat.gob.mx',
        'xd': 'http://www.w3.org/2000/09/xmldsig#'
    }

  def __generar_soapreq__(self, rfc_emisor, rfc_receptor, total, uuid):
     soapreq = ('<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/" xmlns:tem="http://tempuri.org/">'
     + '<soapenv:Header/>'
     +  '<soapenv:Body>'
     +   '<tem:Consulta>'
     +     '<!--Optional:-->'
     +       '<tem:expresionImpresa>'
     +         '<![CDATA[?re='+rfc_emisor+ '&rr=' + rfc_receptor + '&tt=' + total +'&id=' + uuid + ']]>'
     +       '</tem:expresionImpresa>'
     +   '</tem:Consulta>'
     + '</soapenv:Body>'
     +'</soapenv:Envelope>')

     return soapreq

  def obtener_estado(self, rfc_emisor, rfc_receptor, total, uuid):
    soapreq = self.__generar_soapreq__(rfc_emisor, rfc_receptor, total, uuid)

    headers = {
        'Content-type': 'text/xml;charset="utf-8"',
        'Accept': 'text/xml',
        'Cache-Control': 'no-cache',
        'SOAPAction': self.SOAP_ACTION,
    }

    response = requests.post(self.SOAP_URL, data=soapreq, headers=headers, verify=True)


    if response.status_code != requests.codes['ok']:
        if not response.text.startswith('<s:Envelope'):
            ex = 'El webservice Autenticacion responde: {}'.format(response.text)
        else:
            resp_xml = etree.fromstring(response.text)
            ex = resp_xml.find('s:Body/s:Fault/faultstring', namespaces=self.NSMAP).text
        raise Exception(ex)

    if not response.text.startswith('<s:Envelope'):
        ex = 'El webservice Autenticacion responde: {}'.format(response.text)
        raise Exception(ex)

    nsmap = {
        's': 'http://schemas.xmlsoap.org/soap/envelope/',
        't': 'http://tempuri.org/',
        'a': 'http://schemas.datacontract.org/2004/07/Sat.Cfdi.Negocio.ConsultaCfdi.Servicio'
    }

    resp_xml = etree.fromstring(response.text)

    f_val = 's:Body/t:ConsultaResponse/t:ConsultaResult/a:CodigoEstatus'
    CodigoEstatus = resp_xml.find(f_val, namespaces=nsmap)

    f_val = 's:Body/t:ConsultaResponse/t:ConsultaResult/a:EsCancelable'
    EsCancelable = resp_xml.find(f_val, namespaces=nsmap)

    f_val = 's:Body/t:ConsultaResponse/t:ConsultaResult/a:Estado'
    Estado = resp_xml.find(f_val, namespaces=nsmap)

    ret_val = {
        'codigo_estatus': CodigoEstatus.text,
        'es_cancelable': EsCancelable.text,
        'estado': Estado.text
    }
    return ret_val



# -*- coding: utf-8 -*-
import base64
import hashlib
import uuid
from datetime import datetime, timedelta

from .webservicerequest import WebServiceRequest


class Autenticacion(WebServiceRequest):
    DATE_TIME_FORMAT: str = '%Y-%m-%dT%H:%M:%S.%fZ'

    xml_name = 'autenticacion.xml'
    soap_url = 'https://cfdidescargamasivasolicitud.clouda.sat.gob.mx/Autenticacion/Autenticacion.svc'
    soap_action = 'http://DescargaMasivaTerceros.gob.mx/IAutenticacion/Autentica'
    result_xpath = 's:Body/AutenticaResponse/AutenticaResult'
    
    internal_nsmap = {
        's': 'http://schemas.xmlsoap.org/soap/envelope/',
        'o': 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd',
        'u': 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd',
        'des': 'http://DescargaMasivaTerceros.sat.gob.mx',
        '': 'http://www.w3.org/2000/09/xmldsig#',
    }

    external_nsmap = {
        '': 'http://DescargaMasivaTerceros.gob.mx',
        's': 'http://schemas.xmlsoap.org/soap/envelope/',
        'u': 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-utility-1.0.xsd',
        'o': 'http://docs.oasis-open.org/wss/2004/01/oasis-200401-wss-wssecurity-secext-1.0.xsd',
    }

    def obtener_token(self, id=uuid.uuid4(), seconds=300):

        date_created = datetime.utcnow()
        date_expires = date_created + timedelta(seconds=seconds)
        date_created = date_created.strftime(self.DATE_TIME_FORMAT)
        date_expires = date_expires.strftime(self.DATE_TIME_FORMAT)

        self.set_element_text(
            's:Header/o:Security/u:Timestamp/u:Created',
            date_created
        )
        self.set_element_text(
            's:Header/o:Security/u:Timestamp/u:Expires',
            date_expires
        )
        self.set_element_text(
            's:Header/o:Security/o:BinarySecurityToken',
            self.signer.fiel.cer_to_base64(),
        )

        element = self.get_element('s:Header/o:Security/u:Timestamp')
        element_bytes = self.element_to_bytes(element)
        element_hash = hashlib.new('sha1', element_bytes)
        element_digest = element_hash.digest()
        element_digest_base64 = base64.b64encode(element_digest)

        digest_xpath = 's:Header/o:Security/Signature/SignedInfo/Reference/DigestValue'
        self.set_element_text(digest_xpath, element_digest_base64)

        signed_info_xpath = 's:Header/o:Security/Signature/SignedInfo'
        signed_info = self.get_element(signed_info_xpath)
        signed_info_bytes = self.element_to_bytes(signed_info)
        signed_info_sign = self.signer.fiel.firmar_sha1(signed_info_bytes)

        xpath = 's:Header/o:Security/Signature/SignatureValue'
        self.set_element_text(xpath, signed_info_sign)

        element_response = self.request()

        return element_response.text



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='cfdiclient',
    version='1.5.4',
    author='Luis Iturrios',
    author_email='luisiturrios1@gmail.com',
    description='Cliente Python Web Service del SAT para la descarga masiva de CFDIs',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/luisiturrios1/python-cfdiclient',
    packages=setuptools.find_packages(),
    include_package_data=True,
    classifiers=[
        'Programming Language :: Python :: 2.7',
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
    ],
    install_requires=[
        'lxml>=4.2.5',
        'requests>=2.21.0',
        'pycryptodome>=3.7.2',
        'pyOpenSSL>=18.0.0'
    ]
)



# -*- coding: utf-8 -*-
from .webservicerequest import WebServiceRequest


class SolicitaDescarga(WebServiceRequest):

    xml_name = 'solicitadescarga.xml'
    soap_url = 'https://cfdidescargamasivasolicitud.clouda.sat.gob.mx/SolicitaDescargaService.svc'
    soap_action = 'http://DescargaMasivaTerceros.sat.gob.mx/ISolicitaDescargaService/SolicitaDescarga'
    solicitud_xpath = 's:Body/des:SolicitaDescarga/des:solicitud'
    result_xpath = 's:Body/SolicitaDescargaResponse/SolicitaDescargaResult'

    def solicitar_descarga(
        self, token, rfc_solicitante, fecha_inicial, fecha_final,
        rfc_emisor=None, rfc_receptor=None, tipo_solicitud='CFDI',
        tipo_comprobante=None, estado_comprobante=None, 
        rfc_a_cuenta_terceros=None, complemento=None, uuid=None
    ):

        arguments = {
            'RfcSolicitante': rfc_solicitante,
            'FechaFinal': fecha_final.strftime(self.DATE_TIME_FORMAT),
            'FechaInicial': fecha_inicial.strftime(self.DATE_TIME_FORMAT),
            'TipoSolicitud': tipo_solicitud,
            'TipoComprobante': tipo_comprobante,
            'EstadoComprobante': estado_comprobante,
            'RfcACuentaTerceros': rfc_a_cuenta_terceros,
            'Complemento': complemento,
            'UUID': uuid,
        }

        if rfc_emisor:
            arguments['RfcEmisor'] = rfc_emisor

        if rfc_receptor:
            arguments['RfcReceptores'] = [rfc_receptor]

        element_response = self.request(token, arguments)

        ret_val = {
            'id_solicitud': element_response.get('IdSolicitud'),
            'cod_estatus': element_response.get('CodEstatus'),
            'mensaje': element_response.get('Mensaje')
        }

        return ret_val



"""cfdiclient.WebServiceRequest"""
import logging

import requests
from lxml import etree

from .fiel import Fiel
from .signer import Signer
from .utils import Utils

logger = logging.getLogger(__name__)


class WebServiceRequest(Utils):
    """WebServiceRequest
    Base class for signed web service request
    """
    DATE_TIME_FORMAT: str = '%Y-%m-%dT%H:%M:%S'

    soap_url: str = None
    soap_action: str = None
    result_xpath: str = None

    fault_xpath: str = 's:Body/s:Fault/faultstring'

    def __init__(self, fiel: Fiel, verify: bool = True) -> None:
        super().__init__()
        self.signer = Signer(fiel)
        self.verify = verify

    def get_headers(self, token: str) -> dict:
        headers = {
            'Content-type': 'text/xml;charset="utf-8"',
            'Accept': 'text/xml',
            'Cache-Control': 'no-cache',
            'SOAPAction': self.soap_action,
            'Authorization': 'WRAP access_token="{}"'.format(token) if token else ''
        }
        return headers

    def set_request_arguments(self, arguments: dict) -> etree.Element:
        solicitud = self.get_element(self.solicitud_xpath)
        for key in arguments:
            # TODO: Remover esta hardcodeada de aqui
            if key == 'RfcReceptores':
                for i, rfc_receptor in enumerate(arguments[key]):
                    if i == 0:
                        self.set_element_text(
                            's:Body/des:SolicitaDescarga/des:solicitud/des:RfcReceptores/des:RfcReceptor',
                            rfc_receptor
                        )
                        # TODO: Agregar mas de un RFC
                continue
            if arguments[key] != None:
                solicitud.set(key, arguments[key])
        return solicitud

    def request(self, token: str = None, arguments: dict = None) -> etree.Element:

        if arguments:
            solicitud = self.set_request_arguments(arguments)
            self.signer.sign(solicitud)

        headers = self.get_headers(token)

        soap_request = self.element_to_bytes(self.element_root)

        logger.debug('Request soap_url: %s', self.soap_url)
        logger.debug('Request headers: %s', headers)
        logger.debug('Request soap_request: %s', soap_request)

        response = requests.post(
            self.soap_url,
            data=soap_request,
            headers=headers,
            verify=self.verify,
        )

        logger.debug('Response headers: %s', response.headers)
        logger.debug('Response text: %s', response.text)

        try:
            response_xml = etree.fromstring(response.text)
        except Exception:
            raise Exception(response.text)

        if response.status_code != requests.codes['ok']:
            error = self.get_element_external(response_xml, self.fault_xpath)
            raise Exception(error)

        return self.get_element_external(response_xml, self.result_xpath)
